/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <Ecore.h>
#include <hw/thermal.h>

#include "apps/apps.h"
#include "core/devices.h"
#include "core/log.h"
#include "core/device-notifier.h"
#include "core/devices.h"

static struct thermal_device *thermal_dev;
static int noti;

#define ARRAY_SIZE(name) (sizeof(name)/sizeof(name[0]))
#define SIGNAL_OVERHEAT_TIME           "TimeUpdate"
#define OVERHEAT_CALLBACK_TIME		1 //seconds
#define OVERHEAT_POWEROFF_TIME		30

enum thermal_trend_type {
	NO_CHANGE,
	LOW_TO_HIGH,
	HIGH_TO_LOW,
	HIGH_TO_HIGH
};

enum thermal_action_type {
	NOTHING,
	WARNING,
	CRITICAL,
	RECOVER
};

enum thermal_level_update_type {
	UPDATE,
	KEEP
};

struct thermal_state {
	thermal_state_e old_state;
	thermal_state_e new_state;
	enum thermal_action_type status;
};

struct thermal_level {
	thermal_level_e old_level;
	thermal_level_e new_level;
	enum thermal_action_type status;
	enum thermal_level_update_type level;
};

static struct thermal_state temp_state[] = {
	{ THERMAL_STATE_LOW,	THERMAL_STATE_LOW,	NO_CHANGE},
	{ THERMAL_STATE_LOW,	THERMAL_STATE_NORMAL,	NO_CHANGE},
	{ THERMAL_STATE_LOW,	THERMAL_STATE_HIGH,	LOW_TO_HIGH},
	{ THERMAL_STATE_NORMAL, THERMAL_STATE_LOW,	NO_CHANGE},
	{ THERMAL_STATE_NORMAL, THERMAL_STATE_NORMAL,	NO_CHANGE},
	{ THERMAL_STATE_NORMAL, THERMAL_STATE_HIGH,	LOW_TO_HIGH},
	{ THERMAL_STATE_HIGH,	THERMAL_STATE_LOW,	HIGH_TO_LOW},
	{ THERMAL_STATE_HIGH,	THERMAL_STATE_NORMAL,	HIGH_TO_LOW},
	{ THERMAL_STATE_HIGH,	THERMAL_STATE_HIGH,	HIGH_TO_HIGH}
};

static struct thermal_level lowhigh_temp_level[] = {
	{ THERMAL_LEVEL_NORMAL,	THERMAL_LEVEL_WARNING,	WARNING,	UPDATE},
	{ THERMAL_LEVEL_NORMAL,	THERMAL_LEVEL_CRITICAL,	CRITICAL,	UPDATE},
	{ THERMAL_LEVEL_NORMAL,	THERMAL_LEVEL_POWEROFF,	CRITICAL,	UPDATE}
};

static struct thermal_level highlow_temp_level[] = {
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_NORMAL,	NOTHING,	KEEP},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_NORMAL,	NOTHING,	KEEP},
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_NORMAL,	RECOVER,	UPDATE}
};

static struct thermal_level highhigh_temp_level[] = {
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_WARNING,	NOTHING,	UPDATE},
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_CRITICAL,	CRITICAL,	UPDATE},
	{ THERMAL_LEVEL_WARNING,	THERMAL_LEVEL_POWEROFF,	CRITICAL,	UPDATE},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_WARNING,	NOTHING,	KEEP},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_CRITICAL, NOTHING,	UPDATE},
	{ THERMAL_LEVEL_CRITICAL,	THERMAL_LEVEL_POWEROFF, NOTHING,	UPDATE},
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_WARNING,	NOTHING,	KEEP},
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_CRITICAL,	NOTHING,	KEEP},
	{ THERMAL_LEVEL_POWEROFF,	THERMAL_LEVEL_POWEROFF,	NOTHING,	UPDATE}
};

static int power_off(void)
{
	const struct device_ops *power;

	power = find_device("power");
	if (check_default(power))
		return -ENODEV;

	return power->execute("poweroff");
}

static int thermal_state_check(int old_state, int new_state)
{
	int i, ret = -1;

	for (i = 0; i < ARRAY_SIZE(temp_state); i++) {
		if (old_state == temp_state[i].old_state &&
			new_state == temp_state[i].new_state) {
			ret = temp_state[i].status;
			break;
		}
	}

	return ret;
}

static void thermal_up_check(int old_level, int new_level, int *out_status, int *out_level)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(lowhigh_temp_level); i++) {
		if (old_level == lowhigh_temp_level[i].old_level &&
			new_level == lowhigh_temp_level[i].new_level) {
			*out_status = lowhigh_temp_level[i].status;
			*out_level = lowhigh_temp_level[i].level;
			break;
		}
	}
}

static void thermal_down_check(int old_level, int new_level, int *out_status, int *out_level)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(highlow_temp_level); i++) {
		if (old_level == highlow_temp_level[i].old_level &&
			new_level == highlow_temp_level[i].new_level) {
			*out_status = highlow_temp_level[i].status;
			*out_level = highlow_temp_level[i].level;
			break;
		}
	}
}

static void thermal_high_check(int old_level, int new_level, int *out_status, int *out_level)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(highhigh_temp_level); i++) {
		if (old_level == highhigh_temp_level[i].old_level &&
			new_level == highhigh_temp_level[i].new_level) {
			*out_status = highhigh_temp_level[i].status;
			*out_level = highhigh_temp_level[i].level;
			break;
		}
	}
}

static void thermal_remove_noti(void)
{
	if (noti) {
		remove_notification("TempCooldownNotiOff", noti);
		noti = 0;
	}
}

static void thermal_remove_popup(void)
{
	launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "remove_cooldown_popups");
}

static void thermal_add_noti(void)
{
	if (!noti)
		noti = add_notification("TempCooldownNotiOn");
}

static Eina_Bool thermal_overheat_time_broadcast(void *data)
{
	int ret;
	char *arr[1];
	char buff[32];
	static int time = OVERHEAT_POWEROFF_TIME;

	snprintf(buff, sizeof(buff), "%d", time);
	arr[0] = buff;

	/* Transfer Time by Signal */
	ret = broadcast_edbus_signal(POPUP_OVERHEAT_PATH, POPUP_OVERHEAT_INTERFACE,
			SIGNAL_OVERHEAT_TIME, "i", arr);
	if (ret < 0)
		_E("Fail in updating overheat time");

	time -= 1;
	if (time < 0) {
		time = OVERHEAT_POWEROFF_TIME;
		ret = power_off();
		if (ret < 0)
			_E("Fail to power off");
		return ECORE_CALLBACK_CANCEL;
	} else
		return ECORE_CALLBACK_RENEW;
}

static void thermal_add_popup(void)
{
	int ret;
	Ecore_Timer *timer;

	ret = launch_system_app(APP_OVERHEAT, 2,
			APP_KEY_TYPE, "overheat");

	if (ret < 0)
		_E("error to launch Overheat popup");
	else {
		timer = ecore_timer_add(OVERHEAT_CALLBACK_TIME, (Ecore_Task_Cb)thermal_overheat_time_broadcast, NULL);
		if (!timer)
			_E("Fail to set over temp timer");
	}
}

static void thermal_add_recovery_popup(void)
{
	int ret;

	ret = launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "cooled_down");
	if (ret < 0)
		_E("Fail to launch recovery popup");
}

static void thermal_action(struct thermal_info *info, void *data)
{
	static int old_state = -1, old_level = -1;
	static int therm_state = -1, therm_action = -1, update = -1;

	if (!info)
		return;

	if (old_state < 0) {
		old_state = info->state;
		old_level = info->level;
	}

	therm_state = thermal_state_check(old_state, info->state);

	if (therm_state == NO_CHANGE)
		update = UPDATE;
	else if (therm_state == LOW_TO_HIGH) {
		thermal_up_check(old_level, info->level, &therm_action, &update);
		if (therm_action == WARNING) {
		thermal_remove_noti();
		thermal_remove_popup();
		thermal_add_noti();
		} else if (therm_action == CRITICAL) {
			thermal_remove_noti();
			thermal_remove_popup();
			thermal_add_popup();
		}
	} else if (therm_state == HIGH_TO_LOW) {
		thermal_down_check(old_level, info->level, &therm_action, &update);
		if (therm_action == RECOVER) {
			thermal_remove_popup();
			thermal_remove_noti();
			thermal_add_recovery_popup();
		}
	} else if (therm_state == HIGH_TO_HIGH) {
		thermal_high_check(old_level, info->level, &therm_action, &update);
		if (therm_action == CRITICAL) {
			thermal_remove_noti();
			thermal_remove_popup();
			thermal_add_popup();
		}
	}

	if (update == UPDATE) {
		old_state = info->state;
		old_level = info->level;
	}

	_I("Last old_state = %d old_level=%d", old_state, old_level);
}

static int thermal_service_start(void)
{
	int ret;

	if (!thermal_dev) {
		_E("Thermal service is not supported");
		return -ENOTSUP;
	}

	ret = thermal_dev->register_changed_event(thermal_action, NULL);
	if (ret < 0)
		_E("Failed to register thermal event (%d)", ret);

	return ret;
}

static int thermal_service_stop(void)
{
	int ret;

	if (!thermal_dev) {
		_E("Thermal service is not supported");
		return -ENOTSUP;
	}

	ret = thermal_dev->unregister_changed_event(thermal_action);
	if (ret < 0)
		_E("Failed to unregister thermal event (%d)", ret);

	return ret;
}

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		return done;
	done = *(int *)data;
	if (done == 0)
		return done;

	_I("booting done");
	thermal_service_start();

	return done;
}

static void thermal_init(void *data)
{
	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
}

static void thermal_exit(void *data)
{
	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	thermal_service_stop();
}

static int thermal_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (thermal_dev)
		return 0;

	ret = hw_get_info(THERMAL_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("There is no HAL for Thermal");
		return ret;
	}

	if (!info->open) {
		_E("Failed to open thermal device; open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, NULL, (struct hw_common**)&thermal_dev);
	if (ret < 0) {
		_E("Failed to get thermal device structure (%d)", ret);
		return ret;
	}

	if (!thermal_dev ||
		!thermal_dev->get_state ||
		!thermal_dev->register_changed_event ||
		!thermal_dev->unregister_changed_event) {
		_E("Thermal HAL does not support event handler");
		return -ENODEV;
	}

	_I("Thermal device structure load success");
	return 0;
}

static const struct device_ops thermal_ops = {
	.name     = "thermal",
	.probe    = thermal_probe,
	.init     = thermal_init,
	.exit     = thermal_exit,
};

DEVICE_OPS_REGISTER(&thermal_ops)
